#    Copyright (C) 2012-2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from taskflow import exceptions as exc

# Job states.
CLAIMED = 'CLAIMED'
COMPLETE = 'COMPLETE'
UNCLAIMED = 'UNCLAIMED'

# Flow states.
FAILURE = 'FAILURE'
PENDING = 'PENDING'
REVERTING = 'REVERTING'
REVERTED = 'REVERTED'
RUNNING = 'RUNNING'
SUCCESS = 'SUCCESS'
SUSPENDING = 'SUSPENDING'
SUSPENDED = 'SUSPENDED'
RESUMING = 'RESUMING'

# Task states (mainly a subset of the flow states).
FAILURE = FAILURE
PENDING = PENDING
REVERTED = REVERTED
REVERTING = REVERTING
SUCCESS = SUCCESS
RUNNING = RUNNING
RETRYING = 'RETRYING'
IGNORE = 'IGNORE'
REVERT_FAILURE = 'REVERT_FAILURE'

# Atom intentions.
EXECUTE = 'EXECUTE'
IGNORE = IGNORE
REVERT = 'REVERT'
RETRY = 'RETRY'
INTENTIONS = (EXECUTE, IGNORE, REVERT, RETRY)

# Additional engine states
SCHEDULING = 'SCHEDULING'
WAITING = 'WAITING'
ANALYZING = 'ANALYZING'

# Job state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html

_ALLOWED_JOB_TRANSITIONS = frozenset(
    (
        # Job is being claimed.
        (UNCLAIMED, CLAIMED),
        # Job has been lost (or manually unclaimed/abandoned).
        (CLAIMED, UNCLAIMED),
        # Job has been finished.
        (CLAIMED, COMPLETE),
    )
)


def check_job_transition(old_state, new_state):
    """Check that job can transition from from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true. If transition
    should be ignored, it returns false. If transition is not
    valid, it raises an InvalidState exception.
    """
    if old_state == new_state:
        return False
    pair = (old_state, new_state)
    if pair in _ALLOWED_JOB_TRANSITIONS:
        return True
    raise exc.InvalidState(
        "Job transition from '%s' to '%s' is not allowed" % pair
    )


# Flow state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html#flow

_ALLOWED_FLOW_TRANSITIONS = frozenset(
    (
        (PENDING, RUNNING),  # run it!
        (RUNNING, SUCCESS),  # all tasks finished successfully
        (RUNNING, FAILURE),  # some of task failed
        (RUNNING, REVERTED),  # some of task failed and flow has been reverted
        (RUNNING, SUSPENDING),  # engine.suspend was called
        (RUNNING, RESUMING),  # resuming from a previous running
        (SUCCESS, RUNNING),  # see note below
        (FAILURE, RUNNING),  # see note below
        (REVERTED, PENDING),  # try again
        (SUCCESS, PENDING),  # run it again
        (SUSPENDING, SUSPENDED),  # suspend finished
        (SUSPENDING, SUCCESS),  # all tasks finished while we were waiting
        (SUSPENDING, FAILURE),  # some tasks failed while we were waiting
        (
            SUSPENDING,
            REVERTED,
        ),  # all tasks were reverted while we were waiting
        (SUSPENDING, RESUMING),  # resuming from a previous suspending
        (SUSPENDED, RUNNING),  # restart from suspended
        (RESUMING, SUSPENDED),  # after flow resumed, it is suspended
    )
)


# NOTE(imelnikov) SUCCESS->RUNNING and FAILURE->RUNNING transitions are
# useful when flow or flowdetails backing it were altered after the flow
# was finished; then, client code may want to run through flow again
# to ensure all tasks from updated flow had a chance to run.


# NOTE(imelnikov): Engine cannot transition flow from SUSPENDING to
# SUSPENDED while some tasks from the flow are running and some results
# from them are not retrieved and saved properly, so while flow is
# in SUSPENDING state it may wait for some of the tasks to stop. Then,
# flow can go to SUSPENDED, SUCCESS, FAILURE or REVERTED state depending
# of actual state of the tasks -- e.g. if all tasks were finished
# successfully while we were waiting, flow can be transitioned from
# SUSPENDING to SUCCESS state.

_IGNORED_FLOW_TRANSITIONS = frozenset(
    (a, b)
    for a in (PENDING, FAILURE, SUCCESS, SUSPENDED, REVERTED)
    for b in (SUSPENDING, SUSPENDED, RESUMING)
    if a != b
)


def check_flow_transition(old_state, new_state):
    """Check that flow can transition from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true. If transition
    should be ignored, it returns false. If transition is not
    valid, it raises an InvalidState exception.
    """
    if old_state == new_state:
        return False
    pair = (old_state, new_state)
    if pair in _ALLOWED_FLOW_TRANSITIONS:
        return True
    if pair in _IGNORED_FLOW_TRANSITIONS:
        return False
    raise exc.InvalidState(
        "Flow transition from '%s' to '%s' is not allowed" % pair
    )


# Task state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html#task

_ALLOWED_TASK_TRANSITIONS = frozenset(
    (
        (PENDING, RUNNING),  # run it!
        (PENDING, IGNORE),  # skip it!
        (RUNNING, SUCCESS),  # the task executed successfully
        (RUNNING, FAILURE),  # the task execution failed
        (FAILURE, REVERTING),  # task execution failed, try reverting...
        (SUCCESS, REVERTING),  # some other task failed, try reverting...
        (REVERTING, REVERTED),  # the task reverted successfully
        (REVERTING, REVERT_FAILURE),  # the task failed reverting (terminal!)
        (REVERTED, PENDING),  # try again
        (IGNORE, PENDING),  # try again
    )
)


def check_task_transition(old_state, new_state):
    """Check that task can transition from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true, false otherwise.
    """
    pair = (old_state, new_state)
    if pair in _ALLOWED_TASK_TRANSITIONS:
        return True
    return False


# Retry state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html#retry

_ALLOWED_RETRY_TRANSITIONS = list(_ALLOWED_TASK_TRANSITIONS)
_ALLOWED_RETRY_TRANSITIONS.extend(
    [
        (SUCCESS, RETRYING),  # retrying retry controller
        (RETRYING, RUNNING),  # run retry controller that has been retrying
    ]
)
_ALLOWED_RETRY_TRANSITIONS = frozenset(_ALLOWED_RETRY_TRANSITIONS)


def check_retry_transition(old_state, new_state):
    """Check that retry can transition from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true, false otherwise.
    """
    pair = (old_state, new_state)
    if pair in _ALLOWED_RETRY_TRANSITIONS:
        return True
    return False



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy import Table, Column, String, ForeignKey, DateTime, Enum
from sqlalchemy_utils.types import json as json_type

from taskflow.persistence import models
from taskflow import states

Tables = collections.namedtuple(
    'Tables', ['logbooks', 'flowdetails', 'atomdetails']
)

# Column length limits...
NAME_LENGTH = 255
UUID_LENGTH = 64
STATE_LENGTH = 255
VERSION_LENGTH = 64


class JSONType(json_type.JSONType):
    """Customized JSONType using oslo.serialization for json operations"""

    def process_bind_param(self, value, dialect):
        if dialect.name == 'postgresql' and json_type.has_postgres_json:
            return value
        if value is not None:
            value = jsonutils.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if dialect.name == 'postgresql':
            return value
        if value is not None:
            value = jsonutils.loads(value)
        return value


def fetch(metadata):
    """Returns the master set of table objects (which is also there schema)."""
    logbooks = Table(
        'logbooks',
        metadata,
        Column('created_at', DateTime, default=timeutils.utcnow),
        Column('updated_at', DateTime, onupdate=timeutils.utcnow),
        Column('meta', JSONType),
        Column('name', String(length=NAME_LENGTH)),
        Column(
            'uuid',
            String(length=UUID_LENGTH),
            primary_key=True,
            nullable=False,
            unique=True,
            default=uuidutils.generate_uuid,
        ),
    )
    flowdetails = Table(
        'flowdetails',
        metadata,
        Column('created_at', DateTime, default=timeutils.utcnow),
        Column('updated_at', DateTime, onupdate=timeutils.utcnow),
        Column(
            'parent_uuid',
            String(length=UUID_LENGTH),
            ForeignKey('logbooks.uuid', ondelete='CASCADE'),
        ),
        Column('meta', JSONType),
        Column('name', String(length=NAME_LENGTH)),
        Column('state', String(length=STATE_LENGTH)),
        Column(
            'uuid',
            String(length=UUID_LENGTH),
            primary_key=True,
            nullable=False,
            unique=True,
            default=uuidutils.generate_uuid,
        ),
    )
    atomdetails = Table(
        'atomdetails',
        metadata,
        Column('created_at', DateTime, default=timeutils.utcnow),
        Column('updated_at', DateTime, onupdate=timeutils.utcnow),
        Column('meta', JSONType),
        Column(
            'parent_uuid',
            String(length=UUID_LENGTH),
            ForeignKey('flowdetails.uuid', ondelete='CASCADE'),
        ),
        Column('name', String(length=NAME_LENGTH)),
        Column('version', String(length=VERSION_LENGTH)),
        Column('state', String(length=STATE_LENGTH)),
        Column(
            'uuid',
            String(length=UUID_LENGTH),
            primary_key=True,
            nullable=False,
            unique=True,
            default=uuidutils.generate_uuid,
        ),
        Column('failure', JSONType),
        Column('results', JSONType),
        Column('revert_results', JSONType),
        Column('revert_failure', JSONType),
        Column('atom_type', Enum(*models.ATOM_TYPES, name='atom_types')),
        Column('intention', Enum(*states.INTENTIONS, name='intentions')),
    )
    return Tables(logbooks, flowdetails, atomdetails)



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import contextlib
import threading

import futurist
from oslo_utils import uuidutils
import testscenarios
import testtools

from taskflow.conductors import backends
from taskflow import engines
from taskflow.jobs.backends import impl_zookeeper
from taskflow.jobs import base
from taskflow.patterns import linear_flow as lf
from taskflow.persistence.backends import impl_memory
from taskflow import states as st
from taskflow import test
from taskflow.tests import utils as test_utils
from taskflow.utils import kazoo_utils
from taskflow.utils import persistence_utils as pu
from taskflow.utils import threading_utils


TEST_PATH_TPL = '/taskflow/conductor-test/%s'
ZOOKEEPER_AVAILABLE = test_utils.zookeeper_available(
    impl_zookeeper.ZookeeperJobBoard.MIN_ZK_VERSION
)


def test_factory(blowup):
    f = lf.Flow("test")
    if not blowup:
        f.add(test_utils.ProgressingTask('test1'))
    else:
        f.add(test_utils.FailingTask("test1"))
    return f


def sleep_factory():
    f = lf.Flow("test")
    f.add(test_utils.SleepTask('test1'))
    f.add(test_utils.ProgressingTask('test2'))
    return f


def test_store_factory():
    f = lf.Flow("test")
    f.add(test_utils.TaskMultiArg('task1'))
    return f


def single_factory():
    return futurist.ThreadPoolExecutor(max_workers=1)


ComponentBundle = collections.namedtuple(
    'ComponentBundle', ['board', 'persistence', 'conductor']
)


@testtools.skipIf(not ZOOKEEPER_AVAILABLE, 'zookeeper is not available')
class ManyConductorTest(
    testscenarios.TestWithScenarios, test_utils.EngineTestBase, test.TestCase
):
    scenarios = [
        (
            'blocking',
            {'kind': 'blocking', 'conductor_kwargs': {'wait_timeout': 0.1}},
        ),
        (
            'nonblocking_many_thread',
            {'kind': 'nonblocking', 'conductor_kwargs': {'wait_timeout': 0.1}},
        ),
        (
            'nonblocking_one_thread',
            {
                'kind': 'nonblocking',
                'conductor_kwargs': {
                    'executor_factory': single_factory,
                    'wait_timeout': 0.1,
                },
            },
        ),
    ]

    def make_components(self):
        def cleanup_path(client, path):
            if not client.connected:
                return
            client.delete(path, recursive=True)

        client = kazoo_utils.make_client(test_utils.ZK_TEST_CONFIG.copy())
        path = TEST_PATH_TPL % uuidutils.generate_uuid()
        persistence = impl_memory.MemoryBackend()
        board = impl_zookeeper.ZookeeperJobBoard(
            'testing', {'path': path}, client=client, persistence=persistence
        )

        self.addCleanup(cleanup_path, client, path)
        self.addCleanup(kazoo_utils.finalize_client, client)

        conductor_kwargs = self.conductor_kwargs.copy()
        conductor_kwargs['persistence'] = persistence
        conductor = backends.fetch(
            self.kind, 'testing', board, **conductor_kwargs
        )
        return ComponentBundle(board, persistence, conductor)

    def test_connection(self):
        components = self.make_components()
        components.conductor.connect()
        with contextlib.closing(components.conductor):
            self.assertTrue(components.board.connected)
        self.assertFalse(components.board.connected)

    def test_run_empty(self):
        components = self.make_components()
        components.conductor.connect()
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)
            t.join()

    def test_run(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()
        job_consumed_event = threading.Event()
        job_abandoned_event = threading.Event()

        def on_consume(state, details):
            consumed_event.set()

        def on_job_consumed(event, details):
            if event == 'job_consumed':
                job_consumed_event.set()

        def on_job_abandoned(event, details):
            if event == 'job_abandoned':
                job_abandoned_event.set()

        components.board.notifier.register(base.REMOVAL, on_consume)
        components.conductor.notifier.register("job_consumed", on_job_consumed)
        components.conductor.notifier.register(
            "job_abandoned", on_job_abandoned
        )
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            lb, fd = pu.temporary_flow_detail(components.persistence)
            engines.save_factory_details(
                fd, test_factory, [False], {}, backend=components.persistence
            )
            components.board.post('poke', lb, details={'flow_uuid': fd.uuid})
            self.assertTrue(consumed_event.wait(test_utils.WAIT_TIMEOUT))
            self.assertTrue(job_consumed_event.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(job_abandoned_event.wait(1))
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)

        persistence = components.persistence
        with contextlib.closing(persistence.get_connection()) as conn:
            lb = conn.get_logbook(lb.uuid)
            fd = lb.find(fd.uuid)
        self.assertIsNotNone(fd)
        self.assertEqual(st.SUCCESS, fd.state)

    def test_run_max_dispatches(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()

        def on_consume(state, details):
            consumed_event.set()

        components.board.notifier.register(base.REMOVAL, on_consume)
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(
                lambda: components.conductor.run(max_dispatches=5)
            )
            t.start()
            lb, fd = pu.temporary_flow_detail(components.persistence)
            engines.save_factory_details(
                fd, test_factory, [False], {}, backend=components.persistence
            )
            for _ in range(5):
                components.board.post(
                    'poke', lb, details={'flow_uuid': fd.uuid}
                )
                self.assertTrue(consumed_event.wait(test_utils.WAIT_TIMEOUT))
            components.board.post('poke', lb, details={'flow_uuid': fd.uuid})
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)

    def test_fail_run(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()
        job_consumed_event = threading.Event()
        job_abandoned_event = threading.Event()

        def on_consume(state, details):
            consumed_event.set()

        def on_job_consumed(event, details):
            if event == 'job_consumed':
                job_consumed_event.set()

        def on_job_abandoned(event, details):
            if event == 'job_abandoned':
                job_abandoned_event.set()

        components.board.notifier.register(base.REMOVAL, on_consume)
        components.conductor.notifier.register("job_consumed", on_job_consumed)
        components.conductor.notifier.register(
            "job_abandoned", on_job_abandoned
        )
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            lb, fd = pu.temporary_flow_detail(components.persistence)
            engines.save_factory_details(
                fd, test_factory, [True], {}, backend=components.persistence
            )
            components.board.post('poke', lb, details={'flow_uuid': fd.uuid})
            self.assertTrue(consumed_event.wait(test_utils.WAIT_TIMEOUT))
            self.assertTrue(job_consumed_event.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(job_abandoned_event.wait(1))
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)

        persistence = components.persistence
        with contextlib.closing(persistence.get_connection()) as conn:
            lb = conn.get_logbook(lb.uuid)
            fd = lb.find(fd.uuid)
        self.assertIsNotNone(fd)
        self.assertEqual(st.REVERTED, fd.state)

    def test_missing_store(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()

        def on_consume(state, details):
            consumed_event.set()

        components.board.notifier.register(base.REMOVAL, on_consume)
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            lb, fd = pu.temporary_flow_detail(components.persistence)
            engines.save_factory_details(
                fd, test_store_factory, [], {}, backend=components.persistence
            )
            components.board.post('poke', lb, details={'flow_uuid': fd.uuid})
            self.assertTrue(consumed_event.wait(test_utils.WAIT_TIMEOUT))
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)

        persistence = components.persistence
        with contextlib.closing(persistence.get_connection()) as conn:
            lb = conn.get_logbook(lb.uuid)
            fd = lb.find(fd.uuid)
        self.assertIsNotNone(fd)
        self.assertIsNone(fd.state)

    def test_job_store(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()

        def on_consume(state, details):
            consumed_event.set()

        store = {'x': True, 'y': False, 'z': None}

        components.board.notifier.register(base.REMOVAL, on_consume)
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            lb, fd = pu.temporary_flow_detail(components.persistence)
            engines.save_factory_details(
                fd, test_store_factory, [], {}, backend=components.persistence
            )
            components.board.post(
                'poke', lb, details={'flow_uuid': fd.uuid, 'store': store}
            )
            self.assertTrue(consumed_event.wait(test_utils.WAIT_TIMEOUT))
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)

        persistence = components.persistence
        with contextlib.closing(persistence.get_connection()) as conn:
            lb = conn.get_logbook(lb.uuid)
            fd = lb.find(fd.uuid)
        self.assertIsNotNone(fd)
        self.assertEqual(st.SUCCESS, fd.state)

    def test_flowdetails_store(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()

        def on_consume(state, details):
            consumed_event.set()

        store = {'x': True, 'y': False, 'z': None}

        components.board.notifier.register(base.REMOVAL, on_consume)
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            lb, fd = pu.temporary_flow_detail(
                components.persistence, meta={'store': store}
            )
            engines.save_factory_details(
                fd, test_store_factory, [], {}, backend=components.persistence
            )
            components.board.post('poke', lb, details={'flow_uuid': fd.uuid})
            self.assertTrue(consumed_event.wait(test_utils.WAIT_TIMEOUT))
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)

        persistence = components.persistence
        with contextlib.closing(persistence.get_connection()) as conn:
            lb = conn.get_logbook(lb.uuid)
            fd = lb.find(fd.uuid)
        self.assertIsNotNone(fd)
        self.assertEqual(st.SUCCESS, fd.state)

    def test_combined_store(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()

        def on_consume(state, details):
            consumed_event.set()

        flow_store = {'x': True, 'y': False}
        job_store = {'z': None}

        components.board.notifier.register(base.REMOVAL, on_consume)
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            lb, fd = pu.temporary_flow_detail(
                components.persistence, meta={'store': flow_store}
            )
            engines.save_factory_details(
                fd, test_store_factory, [], {}, backend=components.persistence
            )
            components.board.post(
                'poke', lb, details={'flow_uuid': fd.uuid, 'store': job_store}
            )
            self.assertTrue(consumed_event.wait(test_utils.WAIT_TIMEOUT))
            components.conductor.stop()
            self.assertTrue(components.conductor.wait(test_utils.WAIT_TIMEOUT))
            self.assertFalse(components.conductor.dispatching)

        persistence = components.persistence
        with contextlib.closing(persistence.get_connection()) as conn:
            lb = conn.get_logbook(lb.uuid)
            fd = lb.find(fd.uuid)
        self.assertIsNotNone(fd)
        self.assertEqual(st.SUCCESS, fd.state)

    def test_stop_aborts_engine(self):
        components = self.make_components()
        components.conductor.connect()
        consumed_event = threading.Event()
        job_consumed_event = threading.Event()
        job_abandoned_event = threading.Event()
        running_start_event = threading.Event()

        def on_running_start(event, details):
            running_start_event.set()

        def on_consume(state, details):
            consumed_event.set()

        def on_job_consumed(event, details):
            if event == 'job_consumed':
                job_consumed_event.set()

        def on_job_abandoned(event, details):
            if event == 'job_abandoned':
                job_abandoned_event.set()

        components.board.notifier.register(base.REMOVAL, on_consume)
        components.conductor.notifier.register("job_consumed", on_job_consumed)
        components.conductor.notifier.register(
            "job_abandoned", on_job_abandoned
        )
        components.conductor.notifier.register(
            "running_start", on_running_start
        )
        with contextlib.closing(components.conductor):
            t = threading_utils.daemon_thread(components.conductor.run)
            t.start()
            lb, fd = pu.temporary_flow_detail(components.persistence)
            engines.save_factory_details(
                fd, sleep_factory, [], {}, backend=components.persistence
            )
            components.board.post(
                'poke',
                lb,
                details={'flow_uuid': fd.uuid, 'store': {'duration': 2}},
            )
            running_start_event.wait(test_utils.WAIT_TIMEOUT)
            components.conductor.stop()
            job_abandoned_event.wait(test_utils.WAIT_TIMEOUT)
            self.assertTrue(job_abandoned_event.is_set())
            self.assertFalse(job_consumed_event.is_set())
            self.assertFalse(consumed_event.is_set())


@testtools.skipIf(not ZOOKEEPER_AVAILABLE, 'zookeeper is not available')
class NonBlockingExecutorTest(test.TestCase):
    def test_bad_wait_timeout(self):
        persistence = impl_memory.MemoryBackend()
        board = impl_zookeeper.ZookeeperJobBoard(
            'testing',
            test_utils.ZK_TEST_CONFIG.copy(),
            persistence=persistence,
        )
        self.assertRaises(
            ValueError,
            backends.fetch,
            'nonblocking',
            'testing',
            board,
            persistence=persistence,
            wait_timeout='testing',
        )

    def test_bad_factory(self):
        persistence = impl_memory.MemoryBackend()
        board = impl_zookeeper.ZookeeperJobBoard(
            'testing',
            test_utils.ZK_TEST_CONFIG.copy(),
            persistence=persistence,
        )
        self.assertRaises(
            ValueError,
            backends.fetch,
            'nonblocking',
            'testing',
            board,
            persistence=persistence,
            executor_factory='testing',
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import futurist

from taskflow.conductors.backends import impl_executor


class BlockingConductor(impl_executor.ExecutorConductor):
    """Blocking conductor that processes job(s) in a blocking manner."""

    MAX_SIMULTANEOUS_JOBS = 1
    """
    Default maximum number of jobs that can be in progress at the same time.
    """

    @staticmethod
    def _executor_factory():
        return futurist.SynchronousExecutor()

    def __init__(
        self,
        name,
        jobboard,
        persistence=None,
        engine=None,
        engine_options=None,
        wait_timeout=None,
        log=None,
        max_simultaneous_jobs=MAX_SIMULTANEOUS_JOBS,
    ):
        super().__init__(
            name,
            jobboard,
            persistence=persistence,
            engine=engine,
            engine_options=engine_options,
            wait_timeout=wait_timeout,
            log=log,
            max_simultaneous_jobs=max_simultaneous_jobs,
        )
